"""
Objects that can be placed in a label template
"""

from PIL import Image, ImageDraw, ImageFont
import qrcode

class Text:
    """A simple text item"""
    def __init__(self, font: ImageFont = None, **kwargs) -> None:
        if font:
            self.font = font
        else:
            # fallback to default font
            self.font = ImageFont.load_default()

        self.pad_top = kwargs.get("pad_top", 0)
        self.pad_right = kwargs.get("pad_right", 0)
        self.pad_bottom = kwargs.get("pad_bottom", 0)
        self.pad_left = kwargs.get("pad_left", 0)

    def render(self, text):
        text_x, text_y = self.font.getsize(text)
        padded_size = (
            text_x + self.pad_left + self.pad_right,
            text_y + self.pad_top + self.pad_bottom,
        )
        image = Image.new("1", padded_size, "white")
        draw = ImageDraw.Draw(image)
        draw.text((self.pad_left, self.pad_top), text, "black", self.font)
        return image

class QRCode:
    """A simple QRCode item"""
    def __init__(self, version=1, error_correction=qrcode.constants.ERROR_CORRECT_L, box_size=1, border=0) -> None:
        self.qr = qrcode.QRCode(
            version=version,
            error_correction=error_correction,
            box_size=box_size,
            border=border,
        )

    def render(self, data):
        self.qr.add_data(data)

        return self.qr.make_image(fill_color="black", back_color="white")


import usb.core
import usb.util
import threading
import serial

def is_usb_printer(dev):
    if dev.bDeviceClass == 7:
        return True
    for cfg in dev:
        if usb.util.find_descriptor(cfg, bInterfaceClass=7) is not None:
            return True


class PyUSBBackend():
    def __init__(self, dev):
        self.dev = dev
        self.lock = threading.Lock()

    @classmethod
    def auto(cls):
        dev = usb.core.find(custom_match=is_usb_printer)
        if dev is None:
            raise OSError('Device not found')
        return cls(dev)

    def write(self, data: bytes):
        self.dev.write(0x2, data)

    def read(self, count: int) -> bytes:
        return self.dev.read(0x81, count)

class BTSerialBackend():
    def __init__(self, dev):
        self.dev = dev
        self.lock = threading.Lock()

    @classmethod
    def auto(cls, devPath: str):
        dev = serial.Serial(
            devPath,
            baudrate=9600,
            stopbits=serial.STOPBITS_ONE,
            parity=serial.PARITY_NONE,
            bytesize=8,
            dsrdtr=False,
            timeout=1
        )
        if dev is None:
            raise OSError('Device not found')
        return cls(dev)

    def write(self, data: bytes):
        self.dev.write(data)

    def read(self, count: int) -> bytes:
        data = self.dev.read(count)
        return data


"""
Labels are the Base class you derive your Labels from. A few simple Labels are
provided for you.
"""
from typing import Tuple

from PIL import Image


def _coord_add(tup1, tup2):
    """add two tuples of size two"""
    return (tup1[0] + tup2[0], tup1[1] + tup2[1])


class Label:
    """Base class for all labels

    >>> class MyLabel(Label):
    ...     items = [
    ...         [Text(), Text()]
    ...     ]
    >>> l = MyLabel("text1", "text2")
    >>> printer.print(l)

    """
    items = []  # type: list

    def __init__(self, *args):
        if not self.items:
            raise ValueError(
                "A Labels 'items' attribute must contain a list of "
                "renderable objects")

        arg_it = iter(args)
        try:
            self._rendered_items = [
                [item.render(next(arg_it)) for item in line]
                for line in self.items]
        except StopIteration:
            # the argument list was exhausted before all items had a value
            raise TypeError("{cls} requires {argc} arguments, but {num} were given".format(
                cls=self.__class__.__name__, argc=sum(len(x) for x in self.items), num=len(args)
            ))

    @property
    def size(self) -> Tuple[int, int]:
        width = max(sum(i.size[0] for i in line)
                    for line in self._rendered_items)
        height = sum(max(i.size[1] for i in line)
                     for line in self._rendered_items)

        return width, height

    def render(self, width=None, height=None) -> Image:
        """render the Label.

        Args:
            width: Width request
            height: Height request
        """
        size = self.size
        img = Image.new("1", size, "white")

        pos = [0, 0]

        for line in self._rendered_items:
            for item in line:
                box = (*pos, *_coord_add(item.size, pos))
                img.paste(item, box=box)
                pos[0] += item.size[0]

            pos[0] = 0
            pos[1] += max(i.size[1] for i in line)

        xdim, ydim = img.size
        xdim = round((height / ydim) * xdim)

        img = img.resize((xdim, height))

        return img



from setuptools import setup, find_packages

with open('README.md', 'r') as readme:
    README_TEXT = readme.read()

setup(
    name="brother_label_printer",
    version="0.0.6",
    description="A library for creating and printing labels",
    use_scm_version=True,
    long_description=README_TEXT,
    long_description_content_type="text/markdown",
    url="https://github.com/Avi-Singh12/labelprinterkit.git",
    author="Avi Singh",
    author_email="singhavi@umich.edu",
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    license='Apache License, Version 2.0',
    keywords='',
    packages=find_packages(),
    install_requires=[
        'pillow',
        'pyusb',
        'packbits',
        'pyserial',
        'qrcode'
    ],
    python_requires='>=3.4'
)



"""
Brother P-Touch P700 Driver
"""
import io
import random
import struct
import time
from collections import namedtuple
from enum import Enum, IntEnum
from itertools import chain, islice
from pprint import pprint
from typing import Iterable, Sequence
from math import ceil
import logging

import packbits
from PIL import Image, ImageChops

from ..label import Label
from . import BasePrinter, BaseStatus, BaseErrorStatus

logger = logging.getLogger(__name__)


def batch_iter_bytes(b, size):
    i = iter(b)

    return iter(lambda: bytes(tuple(islice(i, size))), b"")


class INFO_OFFSETS(IntEnum):
    PRINTHEAD_MARK = 0
    MODEL_CODE = 4
    ERROR_1 = 8
    ERROR_2 = 9
    MEDIA_WIDTH = 10
    MEDIA_TYPE = 11
    MODE = 15
    MEDIA_LENGTH = 17
    STATUS_TYPE = 18
    PHASE_TYPE = 19
    PHASE_NUMBER_HI = 20
    PHASE_NUMBER_LO = 21
    NOTIFY_NO = 22
    HARDWARE_SETTINGS = 26


class ERRORS(Enum):
    NO_MEDIA = 0
    CUTTER_JAM = 2
    WEAK_BATTERY = 3
    HV_ADAPTER = 6
    REPLACE_MEDIA = 8
    COVER_OPEN = 12
    OVERHEATING = 13
    UNKNOWN = -1


class MEDIA_TYPE(Enum):
    NO_MEDIA = 0
    LAMINATED_TAPE = 1
    NON_LAMINATED_TAPE = 2
    HEAT_SHRINK = 3
    INCOMPATIBLE = 4


class STATUS_TYPE(Enum):
    STATUS_REPLY = 0
    PRINTING_DONE = 1
    ERROR_OCCURRED = 2
    TURNED_OFF = 3
    NOTIFICATION = 4
    PHASE_CHANGE = 5


STATUS_TYPE_MAP = {
    0x00: STATUS_TYPE.STATUS_REPLY,
    0x01: STATUS_TYPE.PRINTING_DONE,
    0x02: STATUS_TYPE.ERROR_OCCURRED,
    0x04: STATUS_TYPE.TURNED_OFF,
    0x05: STATUS_TYPE.NOTIFICATION,
    0x06: STATUS_TYPE.PHASE_CHANGE,
}

MEDIA_TYPE_MAP = {
    0x00: MEDIA_TYPE.NO_MEDIA,
    0x01: MEDIA_TYPE.LAMINATED_TAPE,
    0x02: MEDIA_TYPE.NON_LAMINATED_TAPE,
    0x11: MEDIA_TYPE.HEAT_SHRINK,
    0xFF: MEDIA_TYPE.INCOMPATIBLE,
}

TapeInfo = namedtuple("TapeInfo", ["lmargin", "printarea", "rmargin", "width"])

MEDIA_WIDTH_INFO = {
    # media ID to tape width in dots
    0: TapeInfo(None, None, None, None),
    4: TapeInfo(52, 24, 52, 3.5),
    6: TapeInfo(48, 32, 48, 6.0),
    9: TapeInfo(39, 50, 39, 9.0),
    12: TapeInfo(29, 70, 29, 12.0),
    18: TapeInfo(8, 112, 8, 18.0),
    24: TapeInfo(0, 128, 0, 24.0),
}

ERROR_MASK = {
    0: ERRORS.NO_MEDIA,
    2: ERRORS.CUTTER_JAM,
    3: ERRORS.WEAK_BATTERY,
    6: ERRORS.HV_ADAPTER,
    8: ERRORS.REPLACE_MEDIA,
    12: ERRORS.COVER_OPEN,
    13: ERRORS.OVERHEATING,
}


def encode_line(bitmap_line: bytes, tape_info: TapeInfo) -> bytes:
    # The number of bits we need to add left or right is not always a multiple
    # of 8, so we need to convert our line into an int, shift it over by the
    # left margin and convert it to back again, padding to 16 bytes.

    line_int = int.from_bytes(bitmap_line, byteorder='big')
    line_int <<= tape_info.rmargin
    padded = line_int.to_bytes(16, byteorder='big')

    # pad to 16 bytes
    compressed = packbits.encode(padded)
    logger.debug("original bitmap: %s", bitmap_line)
    logger.debug("padded bitmap %s", padded)
    logger.debug("packbits compressed %s", compressed)
    # <h: big endian short (2 bytes)
    prefix = struct.pack("<H", len(compressed))

    return prefix + compressed


class Errors(BaseErrorStatus):
    def __init__(self, byte1: int, byte2: int) -> None:
        value = byte1 | (byte2 << 8)
        self.data = {
            err.name.lower(): bool(value & 1 << offset)

            for offset, err in ERROR_MASK.items()
        }

    def any(self):
        return any(self.data.values())

    def __getattr__(self, attr):
        return self.data[attr]

    def __repr__(self):
        return "<Errors {}>".format(self.data)


class Status(BaseStatus):
    def __init__(self, msg: Sequence) -> None:
        self.data = {i.name.lower(): msg[i.value] for i in INFO_OFFSETS}

        self.errors = Errors(self.error_1, self.error_2)
        self.tape_info = MEDIA_WIDTH_INFO[self.media_width]

    def ready(self):
        return not self.errors.any()

    def __getattr__(self, attr):
        return self.data[attr]


class P700(BasePrinter):
    """Printer Class for the Brother P-Touch P700/PT-700 Printer"""
    DPI = (180, 180)

    def connect(self) -> None:
        """Connect to Printer"""
        self.io.write(b'\x00' * 100)
        self.io.write(b'\x1b\x40')

        logger.info("connected")

    def get_status(self) -> Status:
        """get status of the printer as ``Status`` object"""
        with self.io.lock:
            self.io.write(b'\x1B\x69\x53')
            data = self.io.read(32)

            if not data:
                raise IOError("No Response from printer")

            if len(data) < 32:
                raise IOError("Invalid Response from printer")

            return Status(data)

    def _debug_status(self):
        data = self.io.read(32)

        if data:
            logger.debug(Status(data))

    def get_label_width(self):
        return self.get_status().tape_info.width

    def print_label(self, label: Label) -> Status:
        status = self.get_status()
        if not status.ready():
            raise IOError("Printer is not ready")

        img = label.render(height=status.tape_info.printarea)
        logger.debug("printarea is %s dots", status.tape_info.printarea)
        if not img.mode == "1":
            raise ValueError("render output has invalid mode '1'")
        img = img.transpose(Image.ROTATE_270).transpose(
            Image.FLIP_TOP_BOTTOM)
        img = ImageChops.invert(img)

        logger.info("label output size: %s", img.size)
        logger.info("tape info: %s", status.tape_info)

        img_bytes = img.tobytes()

        with self.io.lock:
            self._raw_print(
                status, batch_iter_bytes(img_bytes, ceil(img.size[0] / 8)))

        # wait for label to finish printing
        time.sleep(len(img_bytes)/1000 if len(img_bytes)/1000 > 5 else 5)

        return self.get_status()

    def _dummy_print(self, status: Status, document: Iterable[bytes]) -> None:
        for line in document:
            encode_line(line, status.tape_info)

    def _raw_print(self, status: Status, document: Iterable[bytes]) -> None:
        logger.info("starting print")

        self.connect()

        # raster mode
        self.io.write(b'\x1B\x69\x69\x01')

        # Various mode
        self.io.write(b'\x1B\x69\x4D\x40')

        # Advanced mode
        self.io.write(b'\x1B\x69\x4B\x08')

        # margin
        self.io.write(b'\x1B\x69\x64\x0E\x00')

        # Compression mode
        self.io.write(b'\x4D\x02')

        for line in document:
            self.io.write(b'G' + encode_line(line, status.tape_info))

        self.io.write(b'\x5A')

        # end page
        self.io.write(b'\x1A')
        logger.info("end of page")


import io
from abc import ABC, abstractmethod
from typing import Dict, Tuple

from ..label import Label


class BaseErrorStatus(ABC):
    """Represents the errors the printer has"""
    @abstractmethod
    def any(self):
        """return if any errors have occurred"""
        return any(self.data.values())

    def __getattr__(self, attr):
        return self.data[attr]

    def __repr__(self):
        return "<Errors {}>".format(self.data)


class BaseStatus(ABC):
    """Represents the status of the printer"""
    errors = None  # type: BaseErrorStatus

    def __repr__(self):
        return "<Status {} {} {}>".format(self.data, self.errors,
                                          self.tape_info)

    @abstractmethod
    def __getattr__(self, attr):
        pass

    @abstractmethod
    def ready(self) -> bool:
        """return if the printer is ready for printing"""
        pass


class BasePrinter(ABC):
    """Base class for printers. All printers define this API.  Any other
    methods are prefixed with a _ to indicate they are not part of the
    printer API"""

    DPI = None  # type: Tuple[float, float]

    def __init__(self, io_obj: io.BufferedIOBase) -> None:
        self.io = io_obj

    def estimate_label_size(self, label: Label) -> Tuple[float, float]:
        """estimate the Labels size in mm"""

        xdpi, ydpi = self.DPI
        xpixels, ypixels = label.size
        return (xpixels / xdpi) * 2.54, (ypixels / ydpi) * 2.54

    def print_label(self, label: Label) -> BaseStatus:
        """Print the label"""

    @abstractmethod
    def connect(self) -> None:
        """Connect to the Printer"""
        pass
